/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_WASTEDTIMEDATA_HH_
#define HATP_WASTEDTIMEDATA_HH_

#include<iostream>
#include<string>

class WastedTimeData
{
protected:
	/**
	* the name of the human
	*/
	std::string name;

	/**
	* the durations of wasted time in their order of appearence
	*/
	std::vector<double> durations;

	/**
	* the duration of wasted time from the start of the plan
	*/
	double durationFromStart;

	/** 
	* the duration of wasted time to the end of the plan
	*/ 
	double durationToEnd;

public:
	explicit WastedTimeData(std::string const& argName, std::vector<double> const& argDurations, double argFromStart, double argToEnd);
	std::string const& getName() const;
	std::vector<double> const& getDurations() const;
	double getFromStartDuration() const;
	double getToEndDuration() const;
	double getTotalDuration() const;
	~WastedTimeData();
};

/**
 * constructor : set name to argName and durations to argDurations
 */
inline WastedTimeData::WastedTimeData(std::string const& argName, std::vector<double> const& argDurations, double argFromStart, double argToEnd)
	: name(argName), durations(argDurations), durationFromStart(argFromStart), durationToEnd(argToEnd)
{
}

/**
 * return name
 */
inline std::string const& WastedTimeData::getName() const
{
	return name;
}

/**
 * return durations
 */
inline std::vector<double> const& WastedTimeData::getDurations() const
{
	return durations;
}

/** 
 * return fromStartDuration
 */ 
inline double WastedTimeData::getFromStartDuration() const
{ 
	return durationFromStart;
}

/** 
 * return toEndDuration
 */ 
inline double WastedTimeData::getToEndDuration() const
{ 
	return durationToEnd;
}

/**
 * return the sum of members of durations plus durationFromStart and durationToEnd
 */
double WastedTimeData::getTotalDuration() const
{
	double sum=0.;

	for(unsigned int i=0 ; i<durations.size() ; ++i)
		sum += durations[i];

	sum += durationFromStart;
	sum += durationToEnd;

	return sum;
}

/**
 * destructor: empty
 */
inline WastedTimeData::~WastedTimeData()
{
}

#endif
